from rag_colls.processors.file_processor import FileProcessor

processor = FileProcessor()

documents = processor.load_data(file_or_folder_paths=["samples/data/2503.20376v1.pdf"])
not_be_chunked_documents = processor.load_data(
    file_or_folder_paths=["samples/data/2503.20376v1.pdf"],
    should_splits=[False],
)


def test_semantic_chunker():
    from rag_colls.processors.chunkers.semantic_chunker import SemanticChunker

    chunker = SemanticChunker(mocking=True)

    chunked_documents = chunker.chunk(documents)

    assert len(chunked_documents) >= len(documents), (
        "Chunked documents should be more than original documents."
    )

    not_be_chunked_documents_chunked = chunker.chunk(not_be_chunked_documents)

    assert len(not_be_chunked_documents_chunked) == len(not_be_chunked_documents), (
        "Chunked with should_splits False should be equal to original documents."
    )

    first_chunk = chunked_documents[0]
    assert hasattr(first_chunk, "document"), "Chunk does not have document attribute."
    assert hasattr(first_chunk, "metadata"), "Chunk does not have metadata attribute."

    first_not_be_chunked_document = not_be_chunked_documents_chunked[0]
    assert hasattr(first_not_be_chunked_document, "document"), (
        "Chunk does not have document attribute."
    )

    assert hasattr(first_not_be_chunked_document, "metadata"), (
        "Chunk does not have metadata attribute."
    )



from abc import ABC, abstractmethod
from rag_colls.types.core.document import Document


class BaseChunker(ABC):
    @abstractmethod
    def _chunk(self, documents: list[Document], **kwargs) -> list[Document]:
        """
        Chunk the documents.

        Args:
            documents (list[Document]): List of documents to be chunked.
            **kwargs: Additional keyword arguments for the chunking function.

        Returns:
            list[Document]: List of chunked documents.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @abstractmethod
    async def _achunk(self, documents: list[Document], **kwargs) -> list[Document]:
        """
        Asynchronously chunk the documents.

        Args:
            documents (list[Document]): List of documents to be chunked.
            **kwargs: Additional keyword arguments for the chunking function.

        Returns:
            list[Document]: List of chunked documents.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    def chunk(self, documents: list[Document], **kwargs) -> list[Document]:
        """
        Chunk the documents.

        Args:
            documents (list[Document]): List of documents to be chunked.
            **kwargs: Additional keyword arguments for the chunking function.

        Returns:
            list[Document]: List of chunked documents.
        """
        chunked_documents = []
        preprocessed_documents = []
        for doc in documents:
            if not doc.metadata.get("should_split", True):
                chunked_documents.append(doc)
                continue

            preprocessed_documents.append(doc)

        chunked_documents.extend(self._chunk(preprocessed_documents, **kwargs))

        return chunked_documents

    async def achunk(self, documents: list[Document], **kwargs) -> list[Document]:
        """
        Asynchronously chunk the documents.

        Args:
            documents (list[Document]): List of documents to be chunked.
            **kwargs: Additional keyword arguments for the chunking function.

        Returns:
            list[Document]: List of chunked documents.
        """
        chunked_documents = []
        preprocessed_documents = []
        for doc in documents:
            if not doc.metadata.get("should_split", True):
                chunked_documents.append(doc)
                continue

            preprocessed_documents.append(doc)

        chunked_documents.extend(await self._achunk(preprocessed_documents, **kwargs))

        return chunked_documents



import bm25s
import shutil
from pathlib import Path
from loguru import logger

from rag_colls.core.base.database.bm25 import BaseBM25Retriever
from rag_colls.types.retriever import (
    RetrieverResult,
    RetrieverQueryType,
    RetrieverIngestInput,
)


class BM25s(BaseBM25Retriever):
    """
    Wrapper for the BM25s library: `https://github.com/xhluca/bm25s`
    """

    def __init__(self, save_dir: str | Path):
        self.save_dir = save_dir
        self.reloaded_retriever = None

        if Path(self.save_dir).exists():
            self.reloaded_retriever = bm25s.BM25.load(
                save_dir=save_dir, load_corpus=True
            )

    def _test_connection(self):
        """
        Test the connection to the BM25s database.
        """
        # NOTE: Since BM25s is a local database, we don't need to test the connection.
        return True

    def _build_corpus(self, documents: list[RetrieverIngestInput]):
        corpus_json = [
            {"text": doc.document, "metadata": {"id": doc.id, **doc.metadata}}
            for doc in documents
        ]
        return corpus_json

    def _index_documents(self, documents: list[RetrieverIngestInput], **kwargs):
        """
        Ingest documents into the BM25s database.

        Args:
            documents (list[Document]): List of documents to ingest.
        """
        corpus_json = self._build_corpus(documents)
        retriever = bm25s.BM25(corpus=corpus_json)

        corpus_text = [doc["text"] for doc in corpus_json]

        corpus_tokens = bm25s.tokenize(corpus_text)

        retriever.index(corpus_tokens)
        retriever.save(self.save_dir)

    def _add_documents_to_index(self, documents: list[RetrieverIngestInput], **kwargs):
        """
        Add documents to the existing BM25s index by re-indexing the entire corpus.

        Refer to: `https://github.com/xhluca/bm25s/discussions/20`

        Args:
            documents (list[Document]): List of documents to ingest.
        """
        logger.debug("Re-indexing ...")

        reloaded_retriever = bm25s.BM25.load(self.save_dir, load_corpus=True)
        old_corpus = reloaded_retriever.corpus
        new_corpus = self._build_corpus(documents)

        merged_corpus = old_corpus + new_corpus

        shutil.rmtree(self.save_dir, ignore_errors=True)

        self._index_documents(merged_corpus, **kwargs)

    def _add_documents(self, documents: list[RetrieverIngestInput], **kwargs):
        """
        Ingest documents into the BM25s database.

        Args:
            documents (list[Document]): List of documents to ingest.
        """
        if Path(self.save_dir).exists():
            self._add_documents_to_index(documents, **kwargs)
        else:
            self._index_documents(documents, **kwargs)

    def _search(
        self, query: RetrieverQueryType, top_k: int = 5, **kwargs
    ) -> list[RetrieverResult]:
        """
        Perform a search using the BM25s retriever.

        Args:
            query (str): The query string.
            top_k (int): Number of top results to return.

        Returns:
            list[RetrieverResult]: List of retrieved documents with their metadata.
        """
        assert isinstance(query, str), "Query must be a string."

        query_tokens = bm25s.tokenize(query)
        if self.reloaded_retriever is None:
            self.reloaded_retriever = bm25s.BM25.load(self.save_dir, load_corpus=True)

        results, scores = self.reloaded_retriever.retrieve(query_tokens, k=top_k)

        search_results: list[RetrieverResult] = []

        for i in range(results.shape[1]):
            doc, score = results[0, i], scores[0, i]
            search_results.append(
                RetrieverResult(
                    id=doc["metadata"]["id"],
                    score=score,
                    document=doc["text"],
                    metadata=doc["metadata"],
                )
            )

        # Normalize scores to [0, 1]
        max_score = max(scores[0])
        min_score = min(scores[0])

        for result in search_results:
            result.score = (
                (result.score - min_score) / (max_score - min_score)
                if max_score != min_score
                else 0.0
            )

        return search_results



from pathlib import Path

from rag_colls.types.core.document import Document
from rag_colls.core.base.readers.base import BaseReader


class DocxReader(BaseReader):
    def _load_data(
        self,
        file_path: str | Path,
        should_split: bool = True,
        extra_info: dict | None = None,
        encoding: str = "utf-8",
    ) -> list[Document]:
        try:
            import docx
        except ImportError:
            raise ImportError(
                "python-docx is not installed. Please install it with 'pip install python-docx'."
            )

        file_path = Path(file_path)
        if not file_path.exists():
            raise FileNotFoundError(f"File not found: {file_path}")

        if not file_path.suffix.lower() == ".docx":
            raise ValueError(f"File must be a DOCX file: {file_path}")

        # Read the DOCX file
        doc = docx.Document(file_path)
        file_name = file_path.name

        if not extra_info:
            extra_info = {}

        extra_info["file_path"] = str(file_path)
        extra_info["should_split"] = should_split
        extra_info["file_size"] = file_path.stat().st_size
        extra_info["file_type"] = "docx"
        extra_info["source"] = f"{file_name}"

        # TODO: This currently only supports to split each paragraph into a document
        text = []
        for i, paragraph in enumerate(doc.paragraphs):
            if paragraph.text.strip():  # Only add non-empty paragraphs
                text.append(paragraph.text)

        documents = [
            Document(document="\n".join(text).encode(encoding), metadata=extra_info)
        ]

        return documents



from rag_colls.rags.basic_rag import BasicRAG
from rag_colls.llms.litellm_llm import LiteLLM
from rag_colls.embeddings.hf_embedding import HuggingFaceEmbedding
from rag_colls.processors.chunkers.semantic_chunker import SemanticChunker
from rag_colls.databases.vector_databases.chromadb import ChromaVectorDatabase

rag = BasicRAG(
    vector_database=ChromaVectorDatabase(
        persistent_directory="./chroma_db", collection_name="test"
    ),
    chunker=SemanticChunker(embed_model_name="BAAI/bge-base-en-v1.5"),
    llm=LiteLLM(model_name="openai/gpt-4o-mini"),
    embed_model=HuggingFaceEmbedding(model_name="BAAI/bge-base-en-v1.5"),
)

response = rag.search(query="What is Chain Buddy ?", top_k=5)

print(response.content)
print("===========")
print(response.usage)



from pydantic import BaseModel, Field


class Embedding(BaseModel):
    """Base class for all embeddings."""

    embedding: list[float] = Field(
        ..., description="List of floats representing the embedding"
    )
    metadata: dict = Field(
        default_factory=dict, description="Metadata associated with the embedding"
    )

    def _get_metadata(self):
        metadatas = []
        for key, value in self.metadata.items():
            metadatas.append(f"{key}={value}")
        return ", ".join(metadatas)

    def __str__(self):
        metadata_str = self._get_metadata()
        return f"Embedding(dimensions={len(self.embedding)}, {metadata_str})"

    def __repr__(self):
        return self.__str__()



from pathlib import Path
from typing import Any, List, Optional

from rag_colls.types.core.document import Document
from rag_colls.core.base.readers.base import BaseReader


class ExcelReader(BaseReader):
    def __init__(
        self,
        *args: Any,
        pandas_config: Optional[dict] = None,
        row_joiner: str = "\n",
        col_joiner: str = " ",
        **kwargs: Any,
    ) -> None:
        """
        Initialize the ExcelReader with optional pandas configuration and joiners.

        Args:
            pandas_config (dict, optional): Configuration for pandas read_excel function.
            row_joiner (str, optional): Joiner for rows in the output document.
            col_joiner (str, optional): Joiner for columns in the output document.
        """
        super().__init__(*args, **kwargs)
        self._pandas_config = pandas_config or {}
        self._row_joiner = row_joiner if row_joiner else "\n"
        self._col_joiner = col_joiner if col_joiner else " "

    def _load_data(
        self,
        file_path: str | Path,
        should_split: bool = True,
        extra_info: dict | None = None,
        sheet_name: str | int | List[str | int] | None = None,
        encoding: str = "utf-8",
    ) -> list[Document]:
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                "pandas is not installed. Please install it with 'pip install pandas'."
            )

        file_path = Path(file_path)
        if not file_path.exists():
            raise FileNotFoundError(f"File not found: {file_path}")

        if file_path.suffix.lower() not in [".xlsx", ".xls"]:
            raise ValueError(f"File must be an Excel file (.xlsx or .xls): {file_path}")

        # Read the Excel file
        excel_file = pd.ExcelFile(file_path)
        file_name = file_path.name

        if not extra_info:
            extra_info = {}

        extra_info["file_path"] = str(file_path)
        extra_info["should_split"] = should_split
        extra_info["file_size"] = file_path.stat().st_size
        extra_info["file_type"] = "excel"
        extra_info["sheet_names"] = excel_file.sheet_names

        # Handle sheet selection
        if sheet_name is None:
            sheets_to_process = excel_file.sheet_names
        elif isinstance(sheet_name, (str, int)):
            sheets_to_process = [sheet_name]
        else:
            sheets_to_process = sheet_name

        df_sheets = []
        sheet_infos = []
        for sheet in sheets_to_process:
            try:
                temp = []
                df = pd.read_excel(excel_file, sheet_name=sheet)
                # Add sheet-specific metadata
                sheet_extra_info = extra_info.copy()
                sheet_extra_info["source"] = f"{file_name}: Sheet {sheet}"
                sheet_extra_info["num_rows"] = len(df)
                sheet_extra_info["num_cols"] = len(df.columns)
                sheet_extra_info["columns"] = list(df.columns)

                df = df.dropna(axis=0, how="all")
                df = df.fillna("")

                temp.extend(df.values.astype(str).tolist())
                df_sheets.append(temp)
                sheet_infos.append(sheet_extra_info)

            except Exception as e:
                print(f"Warning: Could not process sheet '{sheet}': {str(e)}")
                continue

        outputs = [
            Document(
                document=self._row_joiner.join(
                    [self._col_joiner.join(row) for row in text_list]
                ).encode(encoding),
                metadata=sheet_infos[i],
            )
            for i, text_list in enumerate(df_sheets)
        ]
        return outputs



import inspect
from loguru import logger
from pydantic import BaseModel
from typing import get_type_hints, Type, get_args, get_origin, Callable, Any

from rag_colls.core.constants import DEBUG_MODE


def log_debug(message: str):
    """
    Logs a debug message if DEBUG_MODE is enabled.

    Args:
        message (str): The message to log.
    """
    if DEBUG_MODE:
        logger.debug(message)


def create_function_schema(func: Callable[..., Any]) -> dict:
    """
    Create a schema for a function based on its parameters and return type.
    Used in function calling LLMs.

    Args:
        func (Callable[..., Any]): The function to create a schema for.

    Returns:
        dict: A dictionary representing the schema of the function.
    """
    type_hints = get_type_hints(func)
    params = inspect.signature(func).parameters

    def get_model_properties(model_class: Type[BaseModel]):
        properties = {}
        for field_name, field in model_class.model_fields.items():
            field_info = {"type": "string"}  # Default type

            # Get field type
            field_type = field.annotation
            if get_origin(field_type) is list:
                field_info["type"] = "array"
                item_type = get_args(field_type)[0]
                if isinstance(item_type, type) and issubclass(item_type, BaseModel):
                    field_info["items"] = {
                        "type": "object",
                        "properties": get_model_properties(item_type),
                    }
                else:
                    field_info["items"] = {"type": "string"}  # Default for array items
            elif isinstance(field_type, type) and issubclass(field_type, BaseModel):
                field_info = {
                    "type": "object",
                    "properties": get_model_properties(field_type),
                }
            elif field_type is int:
                field_info["type"] = "integer"
            elif field_type is float:
                field_info["type"] = "number"
            elif field_type is bool:
                field_info["type"] = "boolean"

            # Add description if available
            if field.description:
                field_info["description"] = field.description

            properties[field_name] = field_info
        return properties

    schema = {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": func.__doc__ or f"Get the {func.__name__}",
            "parameters": {"type": "object", "properties": {}, "required": []},
        },
    }

    # Process each parameter
    for param_name, param in params.items():
        param_type = type_hints.get(param_name)

        # Handle Pydantic models
        if isinstance(param_type, type) and issubclass(param_type, BaseModel):
            schema["function"]["parameters"]["properties"][param_name] = {
                "type": "object",
                "properties": get_model_properties(param_type),
            }
        else:
            # Handle basic types
            type_mapping = {
                str: "string",
                int: "integer",
                float: "number",
                bool: "boolean",
                list: "array",
            }
            param_info = {
                "type": type_mapping.get(param_type, "string"),
                "description": f"The {param_name} parameter",
            }

            # Handle lists/arrays
            if get_origin(param_type) is list:
                param_info = {
                    "type": "array",
                    "items": {"type": "string"},  # Default type for array items
                }

            schema["function"]["parameters"]["properties"][param_name] = param_info

        # Add required parameters
        if param.default is inspect.Parameter.empty:
            schema["function"]["parameters"]["required"].append(param_name)

    return schema



from pathlib import Path

from rag_colls.types.core.document import Document
from rag_colls.core.base.readers.base import BaseReader


class TxtReader(BaseReader):
    def _load_data(
        self,
        file_path: str | Path,
        should_split: bool = True,
        extra_info: dict | None = None,
        encoding: str = "utf-8",
    ) -> list[Document]:
        file_path = Path(file_path)
        if not file_path.exists():
            raise FileNotFoundError(f"File not found: {file_path}")

        # Read the text file
        with open(file_path, "r", encoding=encoding) as f:
            text = f.read()
        file_name = file_path.name

        if not extra_info:
            extra_info = {}

        extra_info["file_path"] = str(file_path)
        extra_info["file_size"] = file_path.stat().st_size
        extra_info["file_type"] = "txt"
        extra_info["encoding"] = encoding
        extra_info["should_split"] = should_split

        # Return the entire text as one document
        return [
            Document(
                document=text,
                metadata=dict(
                    source=file_name,
                    **extra_info,
                ),
            )
        ]



from abc import ABC, abstractmethod
from rag_colls.types.reranker import RerankerResult
from rag_colls.types.retriever import RetrieverQueryType, RetrieverResult


class BaseReranker(ABC):
    @abstractmethod
    def _rerank(
        self,
        query: RetrieverQueryType,
        results: list[list[RetrieverResult]],
        top_k: int = 10,
        **kwargs,
    ) -> list[RerankerResult]:
        """
        Rerank the results based on the query.

        Args:
            query (RetrieverQueryType): The query to rerank the results for.
            results (list[list[RetrieverResult]]): The results to rerank.
            top_k (int): The number of top results to return.
            **kwargs: Additional arguments for the reranker.

        Returns:
            list[RetrieverResult]: The reranked results.
        """
        raise NotImplementedError("Rerank method not implemented.")

    def rerank(
        self,
        query: RetrieverQueryType,
        results: list[list[RetrieverResult]],
        top_k: int = 10,
        **kwargs,
    ) -> list[RerankerResult]:
        """
        Rerank the results based on the query.

        Args:
            query (RetrieverQueryType): The query to rerank the results for.
            results (list[list[RetrieverResult]]): The results to rerank.
            top_k (int): The number of top results to return.

        Returns:
            list[RetrieverResult]: The reranked results.
        """
        return self._rerank(query=query, results=results, top_k=top_k, **kwargs)



from loguru import logger
from abc import abstractmethod

from rag_colls.types.retriever import RetrieverResult, RetrieverIngestInput
from rag_colls.core.base.retrievers.base_retriever_provider import BaseRetrieverProvider


class BaseVectorDatabase(BaseRetrieverProvider):
    """
    Abstract base class for vector databases.
    """

    @abstractmethod
    def _check_collection_exists(self, collection_name: str) -> bool:
        """
        Check the connection to the vector database.

        Returns:
            bool: True if the connection is successful, False otherwise.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @abstractmethod
    def _create_collection(self, collection_name: str, **kwargs):
        """
        Create a collection in the vector database.

        Args:
            collection_name (str): The name of the collection to be created.
            **kwargs: Additional keyword arguments for the collection creation.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @abstractmethod
    def _add_documents(self, documents: list[RetrieverIngestInput], **kwargs):
        """
        Add documents to the vector database.

        Args:
            documents (list[RetrieverIngestInput]): List of documents to be added.
            **kwargs: Additional keyword arguments for the add operation.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @abstractmethod
    def _delete_documents(self, document_ids: list[str], **kwargs):
        """
        Delete documents from the vector database.

        Args:
            document_ids (list[str]): List of document IDs to be deleted.
            **kwargs: Additional keyword arguments for the delete operation.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @abstractmethod
    def _search(
        self, query_embedding: list[float], top_k: int = 5, **kwargs
    ) -> list[RetrieverResult]:
        """
        Retrieve documents from the vector database based on the query.

        Args:
            query_embedding (list[float]): The embedding of the query to search for.
            top_k (int): The number of top results to retrieve.
            **kwargs: Additional keyword arguments for the retrieval operation.

        Returns:
            list[RetrieverResult]: List of retrieved documents with their metadata.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @abstractmethod
    async def _asearch(
        self, query_embedding: list[float], top_k: int = 5, **kwargs
    ) -> list[RetrieverResult]:
        """
        Asynchronously retrieve documents from the vector database based on the query.

        Args:
            query_embedding (list[float]): The embedding of the query to search for.
            top_k (int): The number of top results to retrieve.
            **kwargs: Additional keyword arguments for the retrieval operation.

        Returns:
            list[RetrieverResult]: List of retrieved documents with their metadata.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    def search(
        self, query_embedding: list[float], top_k: int = 5, **kwargs
    ) -> list[RetrieverResult]:
        """
        Retrieve documents from the vector database based on the query.

        Args:
            query_embedding (list[float]): The embedding of the query to search for.
            top_k (int): The number of top results to retrieve.
            **kwargs: Additional keyword arguments for the retrieval operation.

        Returns:
            list[RetrieverResult]: List of retrieved documents with their metadata.
        """
        return self._search(query_embedding, top_k, **kwargs)

    async def asearch(
        self, query_embedding: list[float], top_k: int = 5, **kwargs
    ) -> list[RetrieverResult]:
        """
        Asynchronously retrieve documents from the vector database based on the query.

        Args:
            query_embedding (list[float]): The embedding of the query to search for.
            top_k (int): The number of top results to retrieve.
            **kwargs: Additional keyword arguments for the retrieval operation.

        Returns:
            list[RetrieverResult]: List of retrieved documents with their metadata.
        """
        return await self._asearch(query_embedding, top_k, **kwargs)

    def add_documents(self, documents: list[RetrieverIngestInput], **kwargs):
        """
        Add documents to the vector database.

        Args:
            documents (list[RetrieverIngestInput]): List of documents to be added.
            **kwargs: Additional keyword arguments for the add operation.
        """
        if not self._check_collection_exists(self.collection_name):
            logger.warning(
                f"Collection {self.collection_name} does not exist. Creating collection."
            )
            self._create_collection(self.collection_name, **kwargs)

        self._add_documents(documents, **kwargs)

    def delete_documents(self, document_ids: list[str], **kwargs):
        """
        Delete documents from the vector database.

        Args:
            document_ids (list[str]): List of document IDs to be deleted.
            **kwargs: Additional keyword arguments for the delete operation.
        """
        if not self._check_collection_exists(self.collection_name):
            logger.warning(
                f"Collection {self.collection_name} does not exist. Cannot delete documents."
            )
            return

        self._delete_documents(document_ids, **kwargs)



from uuid import uuid4
from pydantic import BaseModel, Field


def generate_uuid_str() -> str:
    """Generate a new string of UUID."""
    return str(uuid4())


class Document(BaseModel):
    """
    Document class to be used in the RAG pipeline.
    """

    id: str = Field(
        default_factory=generate_uuid_str,
        description="Unique identifier for the document.",
    )
    document: str = Field(..., description="The content of the document.")
    metadata: dict = Field(
        default_factory=dict, description="Metadata associated with the document."
    )

    def __str__(self):
        return f"Document(id={self.id}, metadata={self.metadata}), document={self.document}"

    def __repr__(self):
        return self.__str__()



from loguru import logger
from pydantic import BaseModel, Field, ConfigDict

from rag_colls.core.base.llms.base import BaseCompletionLLM
from rag_colls.core.base.embeddings.base import BaseEmbedding

from rag_colls.llms.litellm_llm import LiteLLM
from rag_colls.embeddings.openai_embedding import OpenAIEmbedding


class RagCollsSettings(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    embed_model: BaseEmbedding = Field(
        ..., description="Embedding model to use in the application"
    )
    completion_llm: BaseCompletionLLM = Field(
        ..., description="Completion LLM to use in the application"
    )

    def __str__(self):
        return f"RagCollsSettings(embed_model={self.embed_model}, completion_llm={self.completion_llm})"


GlobalSettings = RagCollsSettings(
    embed_model=OpenAIEmbedding(), completion_llm=LiteLLM()
)

logger.info("GlobalSettings initialized ...")
print(GlobalSettings)



import re


def check_placeholders(template: str, placeholders: list[str]) -> bool:
    """
    Check if all placeholders are present in the template.

    Args:
        template (str): The template string.
        placeholders (list[str]): A list of placeholders to check for.

    Returns:
        bool: True if all placeholders are present, False otherwise.
    """
    for placeholder in placeholders:
        pattern = r"\{" + re.escape(placeholder) + r"\}"
        if not re.search(pattern, template):
            return False
    return True


def extract_placeholders(template: str) -> list[str]:
    """
    Extract all placeholders from the template.

    Args:
        template (str): The template string.

    Returns:
        list[str]: A list of extracted placeholders.
    """
    return re.findall(r"\{(.*?)\}", template)



CONTEXTUAL_PROMPT = """<document>
{WHOLE_DOCUMENT}
</document>
Here is the chunk we want to situate within the whole document
<chunk>
{CHUNK_CONTENT}
</chunk>
Please give a short succinct context to situate this chunk within the overall document for the purposes of improving search retrieval of the chunk. Answer only with the succinct context and nothing else."""



from tqdm import tqdm
from openai import OpenAI
from dotenv import load_dotenv

from rag_colls.types.embedding import Embedding
from rag_colls.types.core.document import Document
from rag_colls.core.base.embeddings.base import BaseEmbedding
from rag_colls.core.constants import (
    OPENAI_EMBEDDING_MODELS,
    DEFAULT_OPENAI_EMBEDDING_MODEL,
)

load_dotenv()


class OpenAIEmbedding(BaseEmbedding):
    client = OpenAI()

    def __init__(self, model_name: str | None = None):
        """
        Initialize the OpenAI embedding model.

        Args:
            model_name (str): The name of the OpenAI model to use. Can only be chosen from
                `[text-embedding-3-small, text-embedding-3-large, text-embedding-ada-002]`.
        """
        if not model_name:
            model_name = DEFAULT_OPENAI_EMBEDDING_MODEL

        assert model_name in OPENAI_EMBEDDING_MODELS, (
            f"Model {model_name} is not supported."
        )

        self.model_name = model_name

    def __str__(self):
        return f"OpenAIEmbedding(model_name={self.model_name})"

    def _get_query_embedding(self, query: str) -> Embedding:
        """
        Returns the embedding of the query.

        Args:
            query (str): The query to be embedded.

        Returns:
            Embedding: The embedding object of the query.
        """
        embedding = self.client.embeddings.create(model=self.model_name, input=query)
        return Embedding(
            embedding=embedding.data[0].embedding,
            metadata={
                "prompt_tokens": embedding.usage.prompt_tokens,
                "total_tokens": embedding.usage.total_tokens,
            },
        )

    def _get_document_embedding(self, document: Document) -> Embedding:
        """
        Returns the embedding of the document.
        Args:
            document (Document): The document to be embedded.

        Returns:
            Embedding: The embedding object of the document.
        """
        embedding = self.client.embeddings.create(
            model=self.model_name, input=document.document
        )
        return Embedding(
            embedding=embedding.data[0].embedding,
            metadata={
                "prompt_tokens": embedding.usage.prompt_tokens,
                "total_tokens": embedding.usage.total_tokens,
            },
        )

    def _get_batch_query_embedding(
        self, queries: list[str], **kwargs
    ) -> list[Embedding]:
        """
        Returns the embeddings of the queries.
        Args:
            queries (list[str]): The list of queries to be embedded.
            **kwargs: Additional keyword arguments for the embedding function.
        Returns:
            list[Embedding]: The list of embedding objects of the queries.
        """
        batch_size = kwargs.get("batch_size", 1)

        embeddings = []
        bar = tqdm(range(0, len(queries), batch_size), desc="Embedding ...")
        for i in bar:
            batch = queries[i : i + batch_size]
            embedding = self.client.embeddings.create(
                model=self.model_name, input=batch
            )
            for emb in embedding.data:
                embeddings.append(
                    Embedding(
                        embedding=emb.embedding,
                        metadata={
                            "prompt_tokens": embedding.usage.prompt_tokens,
                            "total_tokens": embedding.usage.total_tokens,
                        },
                    )
                )
            bar.update(n=len(batch))

        return embeddings

    def _get_batch_document_embedding(
        self, documents: list[Document], **kwargs
    ) -> list[Embedding]:
        """
        Returns the embeddings of the documents.
        Args:
            documents (list[Document]): The list of documents to be embedded.
            **kwargs: Additional keyword arguments for the embedding function.
        Returns:
            list[Embedding]: The list of embedding objects of the documents.
        """
        contents = [doc.document for doc in documents]

        batch_size = kwargs.get("batch_size", 1)
        embeddings = []

        bar = tqdm(range(0, len(contents), batch_size), desc="Embedding ...")
        for i in bar:
            batch = contents[i : i + batch_size]
            embedding = self.client.embeddings.create(
                model=self.model_name, input=batch
            )
            for emb in embedding.data:
                embeddings.append(
                    Embedding(
                        embedding=emb.embedding,
                        metadata={
                            "prompt_tokens": embedding.usage.prompt_tokens,
                            "total_tokens": embedding.usage.total_tokens,
                        },
                    )
                )
            bar.update(n=len(batch))

        return [
            Embedding(
                embedding=emb.embedding,
                metadata={
                    "prompt_tokens": embedding.usage.prompt_tokens,
                    "total_tokens": embedding.usage.total_tokens,
                },
            )
            for emb in embeddings
        ]



from abc import ABC, abstractmethod
from rag_colls.types.llm import Message
from rag_colls.types.llm import LLMOutput


class BaseCompletionLLM(ABC):
    @abstractmethod
    def _complete(self, messages: list[Message], **kwargs) -> LLMOutput:
        """
        Generates a completion based on the provided messages.

        Args:
            messages (list[Message]): List of messages to be sent to the model.
            **kwargs: Additional keyword arguments for the completion function.

        Returns:
            LLMOutput: The output of the model containing the generated content and usage information.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @abstractmethod
    async def _acomplete(self, messages: list[Message], **kwargs) -> LLMOutput:
        """
        Asynchronously generates a completion based on the provided messages.

         Args:
            messages (list[Message]): List of messages to be sent to the model.
            **kwargs: Additional keyword arguments for the completion function.

        Returns:
            LLMOutput: The output of the model containing the generated content and usage information.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    def complete(self, messages: list[Message], **kwargs) -> LLMOutput:
        """
        Generates a completion based on the provided messages.

        Args:
            messages (list[Message]): List of messages to be sent to the model.
            **kwargs: Additional keyword arguments for the completion function.

        Returns:
            str: The generated completion.
        """
        return self._complete(messages, **kwargs)

    async def acomplete(self, messages: list[Message], **kwargs) -> LLMOutput:
        """
        Asynchronously generates a completion based on the provided messages.

        Args:
            messages (list[Message]): List of messages to be sent to the model.
            **kwargs: Additional keyword arguments for the completion function.

        Returns:
            str: The generated completion.
        """
        return await self._acomplete(messages, **kwargs)



import os

DEBUG_MODE = os.getenv("DEBUG", "False").lower() in ["true", "1"]

OPENAI_EMBEDDING_MODELS = [
    "text-embedding-3-small",
    "text-embedding-3-large",
    "text-embedding-ada-002",
]

DEFAULT_OPENAI_EMBEDDING_MODEL = "text-embedding-ada-002"

HF_EMBEDDING_MODELS = [
    "BAAI/bge-small-en-v1.5",
    "BAAI/bge-base-en-v1.5",
    "BAAI/bge-large-en-v1.5",
]

DEFAULT_HF_EMBEDDING_MODEL = "BAAI/bge-base-en-v1.5"

DEFAULT_OPENAI_MODEL = "openai/gpt-4o-mini"



from rag_colls.rags.basic_rag import BasicRAG
from rag_colls.llms.litellm_llm import LiteLLM
from rag_colls.embeddings.openai_embedding import OpenAIEmbedding
from rag_colls.processors.chunkers.semantic_chunker import SemanticChunker
from rag_colls.databases.vector_databases.chromadb import ChromaVectorDatabase

rag = BasicRAG(
    vector_database=ChromaVectorDatabase(
        persistent_directory="./chroma_db", collection_name="test"
    ),
    chunker=SemanticChunker(embed_model_name="text-embedding-ada-002"),
    llm=LiteLLM(model_name="openai/gpt-4o-mini"),
    embed_model=OpenAIEmbedding(model_name="text-embedding-ada-002"),
)

rag.ingest_db(
    file_or_folder_paths=["samples/papers/2409.13588v1.pdf"], batch_embedding=100
)



# REF: https://github.com/run-llama/llama_index/blob/main/llama-index-integrations/readers/llama-index-readers-json/llama_index/readers/json/base.py
"""JSON Reader."""

import json
import re
from typing import Any, Dict, Generator, List, Optional
from pathlib import Path

from rag_colls.core.base.readers.base import BaseReader
from rag_colls.types.core.document import Document


def _depth_first_yield(
    json_data: Any,
    levels_back: int,
    collapse_length: Optional[int],
    path: List[str],
    ensure_ascii: bool = False,
) -> Generator[str, None, None]:
    """Do depth first yield of all of the leaf nodes of a JSON.

    Combines keys in the JSON tree using spaces.

    If levels_back is set to 0, prints all levels.
    If collapse_length is not None and the json_data is <= that number
      of characters, then we collapse it into one line.

    """
    if isinstance(json_data, (dict, list)):
        # only try to collapse if we're not at a leaf node
        json_str = json.dumps(json_data, ensure_ascii=ensure_ascii)
        if collapse_length is not None and len(json_str) <= collapse_length:
            new_path = path[-levels_back:]
            new_path.append(json_str)
            yield " ".join(new_path)
            return
        elif isinstance(json_data, dict):
            for key, value in json_data.items():
                new_path = path[:]
                new_path.append(key)
                yield from _depth_first_yield(
                    value, levels_back, collapse_length, new_path
                )
        elif isinstance(json_data, list):
            for _, value in enumerate(json_data):
                yield from _depth_first_yield(value, levels_back, collapse_length, path)
    else:
        new_path = path[-levels_back:]
        new_path.append(str(json_data))
        yield " ".join(new_path)


class JSONReader(BaseReader):
    """JSON reader.

    Reads JSON documents with options to help us out relationships between nodes.

    Args:
        levels_back (int): the number of levels to go back in the JSON tree, 0
          if you want all levels. If levels_back is None, then we just format the
          JSON and make each line an embedding

        collapse_length (int): the maximum number of characters a JSON fragment
          would be collapsed in the output (levels_back needs to be not None)
          ex: if collapse_length = 10, and
          input is {a: [1, 2, 3], b: {"hello": "world", "foo": "bar"}}
          then a would be collapsed into one line, while b would not.
          Recommend starting around 100 and then adjusting from there.

        is_jsonl (Optional[bool]): If True, indicates that the file is in JSONL format.
        Defaults to False.

        clean_json (Optional[bool]): If True, lines containing only JSON structure are removed.
        This removes lines that are not as useful. If False, no lines are removed and the document maintains a valid JSON object structure.
        If levels_back is set the json is not cleaned and this option is ignored.
        Defaults to True.
    """

    def __init__(
        self,
        levels_back: Optional[int] = None,
        collapse_length: Optional[int] = None,
        ensure_ascii: bool = False,
        is_jsonl: Optional[bool] = False,
        clean_json: Optional[bool] = True,
    ) -> None:
        """Initialize with arguments."""
        super().__init__()
        self.levels_back = levels_back
        self.collapse_length = collapse_length
        self.ensure_ascii = ensure_ascii
        self.is_jsonl = is_jsonl
        self.clean_json = clean_json

    def _load_data(
        self,
        input_file: str,
        should_split: bool = True,
        extra_info: Optional[Dict] = {},
        encoding: str = "utf-8",
    ) -> List[Document]:
        """Load data from the input file."""
        with open(input_file, encoding=encoding) as f:
            load_data = []
            if self.is_jsonl:
                for line in f:
                    load_data.append(json.loads(line.strip()))
            else:
                load_data = [json.load(f)]

            extra_info = extra_info or {}
            extra_info["source"] = f"{Path(input_file).name}"
            extra_info["file_size"] = Path(input_file).stat().st_size
            extra_info["file_path"] = str(Path(input_file))
            extra_info["file_type"] = "json"
            extra_info["should_split"] = should_split
            extra_info["encoding"] = encoding

            documents = []
            for data in load_data:
                if self.levels_back is None and self.clean_json is True:
                    # If levels_back isn't set and clean json is set,
                    # remove lines containing only formatting, we just format and make each
                    # line an embedding
                    json_output = json.dumps(
                        data, indent=0, ensure_ascii=self.ensure_ascii
                    )
                    lines = json_output.split("\n")
                    useful_lines = [
                        line for line in lines if not re.match(r"^[{}\[\],]*$", line)
                    ]
                    documents.append(
                        Document(
                            document="\n".join(useful_lines).encode(encoding),
                            metadata=extra_info,
                        )
                    )

                elif self.levels_back is None and self.clean_json is False:
                    # If levels_back isn't set  and clean json is False, create documents without cleaning
                    json_output = json.dumps(data, ensure_ascii=self.ensure_ascii)
                    documents.append(
                        Document(
                            document=json_output.encode(encoding), metadata=extra_info
                        )
                    )

                elif self.levels_back is not None:
                    # If levels_back is set, we make the embeddings contain the labels
                    # from further up the JSON tree
                    lines = [
                        *_depth_first_yield(
                            data,
                            self.levels_back,
                            self.collapse_length,
                            [],
                            self.ensure_ascii,
                        )
                    ]
                    documents.append(
                        Document(
                            document="\n".join(lines).encode(encoding),
                            metadata=extra_info,
                        )
                    )
            return documents



from tqdm import tqdm
from transformers import AutoTokenizer, AutoModel
import torch
from dotenv import load_dotenv

from rag_colls.types.embedding import Embedding
from rag_colls.types.core.document import Document
from rag_colls.core.base.embeddings.base import BaseEmbedding
from rag_colls.core.constants import (
    HF_EMBEDDING_MODELS,
    DEFAULT_HF_EMBEDDING_MODEL,
)

load_dotenv()


class HuggingFaceEmbedding(BaseEmbedding):
    model = None
    tokenizer = None

    def __init__(self, model_name: str | None = None):
        """
        Initialize the HuggingFace embedding model.

        Args:
            model_name (str): The name of the HuggingFace model to use. Can only be chosen from
                `[sentence-transformers/all-MiniLM-L6-v2, sentence-transformers/all-mpnet-base-v2, sentence-transformers/all-distilroberta-v1]`.
        """
        if not model_name:
            model_name = DEFAULT_HF_EMBEDDING_MODEL

        assert model_name in HF_EMBEDDING_MODELS, (
            f"Model {model_name} is not supported."
        )

        self.model_name = model_name
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModel.from_pretrained(model_name)
        self.model.eval()  # Set model to evaluation mode

    def __str__(self):
        return f"HFEmbedding(model_name={self.model_name})"

    def _mean_pooling(self, model_output, attention_mask):
        """
        Mean pooling of token embeddings to get sentence embeddings.
        """
        token_embeddings = model_output[
            0
        ]  # First element of model_output contains all token embeddings
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        )
        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
            input_mask_expanded.sum(1), min=1e-9
        )

    def _get_query_embedding(self, query: str) -> Embedding:
        """
        Returns the embedding of the query.

        Args:
            query (str): The query to be embedded.

        Returns:
            Embedding: The embedding object of the query.
        """
        # Tokenize the query
        encoded_input = self.tokenizer(
            query, padding=True, truncation=True, return_tensors="pt"
        )

        # Generate embeddings
        with torch.no_grad():
            model_output = self.model(**encoded_input)

        # Perform pooling
        embedding = self._mean_pooling(model_output, encoded_input["attention_mask"])

        # Get token counts
        tokens = self.tokenizer.encode(query, add_special_tokens=False)
        total_tokens = self.tokenizer.encode(query, add_special_tokens=True)

        return Embedding(
            embedding=embedding[0].tolist(),
            metadata={
                "prompt_tokens": len(tokens),
                "total_tokens": len(total_tokens),
            },
        )

    def _get_document_embedding(self, document: Document) -> Embedding:
        """
        Returns the embedding of the document.
        Args:
            document (Document): The document to be embedded.

        Returns:
            Embedding: The embedding object of the document.
        """
        # Tokenize the document
        encoded_input = self.tokenizer(
            document.document, padding=True, truncation=True, return_tensors="pt"
        )

        # Generate embeddings
        with torch.no_grad():
            model_output = self.model(**encoded_input)

        # Perform pooling
        embedding = self._mean_pooling(model_output, encoded_input["attention_mask"])

        # Get token counts
        tokens = self.tokenizer.encode(document.document, add_special_tokens=False)
        total_tokens = self.tokenizer.encode(document.document, add_special_tokens=True)

        return Embedding(
            embedding=embedding[0].tolist(),
            metadata={
                "prompt_tokens": len(tokens),
                "total_tokens": len(total_tokens),
            },
        )

    def _get_batch_query_embedding(
        self, queries: list[str], **kwargs
    ) -> list[Embedding]:
        """
        Returns the embeddings of the queries.
        Args:
            queries (list[str]): The list of queries to be embedded.
            **kwargs: Additional keyword arguments for the embedding function.
        Returns:
            list[Embedding]: The list of embedding objects of the queries.
        """
        batch_size = kwargs.get("batch_size", 1)

        embeddings = []
        bar = tqdm(range(0, len(queries), batch_size), desc="Embedding ...")
        for i in bar:
            batch = queries[i : i + batch_size]

            # Tokenize the batch
            encoded_input = self.tokenizer(
                batch, padding=True, truncation=True, return_tensors="pt"
            )

            # Generate embeddings
            with torch.no_grad():
                model_output = self.model(**encoded_input)

            # Perform pooling
            batch_embeddings = self._mean_pooling(
                model_output, encoded_input["attention_mask"]
            )

            # Process each query in the batch
            for query, emb in zip(batch, batch_embeddings):
                tokens = self.tokenizer.encode(query, add_special_tokens=False)
                total_tokens = self.tokenizer.encode(query, add_special_tokens=True)
                embeddings.append(
                    Embedding(
                        embedding=emb.tolist(),
                        metadata={
                            "prompt_tokens": len(tokens),
                            "total_tokens": len(total_tokens),
                        },
                    )
                )
            bar.update(n=len(batch))

        return embeddings

    def _get_batch_document_embedding(
        self, documents: list[Document], **kwargs
    ) -> list[Embedding]:
        """
        Returns the embeddings of the documents.
        Args:
            documents (list[Document]): The list of documents to be embedded.
            **kwargs: Additional keyword arguments for the embedding function.
        Returns:
            list[Embedding]: The list of embedding objects of the documents.
        """
        contents = [doc.document for doc in documents]

        batch_size = kwargs.get("batch_size", 1)
        embeddings = []

        bar = tqdm(range(0, len(contents), batch_size), desc="Embedding ...")
        for i in bar:
            batch = contents[i : i + batch_size]

            # Tokenize the batch
            encoded_input = self.tokenizer(
                batch, padding=True, truncation=True, return_tensors="pt"
            )

            # Generate embeddings
            with torch.no_grad():
                model_output = self.model(**encoded_input)

            # Perform pooling
            batch_embeddings = self._mean_pooling(
                model_output, encoded_input["attention_mask"]
            )

            # Process each document in the batch
            for doc, emb in zip(batch, batch_embeddings):
                tokens = self.tokenizer.encode(doc, add_special_tokens=False)
                total_tokens = self.tokenizer.encode(doc, add_special_tokens=True)
                embeddings.append(
                    Embedding(
                        embedding=emb.tolist(),
                        metadata={
                            "prompt_tokens": len(tokens),
                            "total_tokens": len(total_tokens),
                        },
                    )
                )
            bar.update(n=len(batch))

        return embeddings



from pydantic import BaseModel, Field


class Message(BaseModel):
    role: str = Field(
        ..., description="Role of the message sender (user, assistant, system)."
    )
    content: str = Field(..., description="Content of the message.")

    def __str__(self):
        return f"Message(role={self.role}, content={self.content})"

    def __repr__(self):
        return self.__str__()


class LLMUsage(BaseModel):
    """
    Class to represent the usage of LLM resources.
    """

    prompt_tokens: int = Field(0, description="Number of tokens in the prompt.")
    completion_tokens: int = Field(0, description="Number of tokens in the completion.")
    total_tokens: int = Field(0, description="Total number of tokens used.")

    def __str__(self):
        return f"LLMUsage(prompt_tokens={self.prompt_tokens}, completion_tokens={self.completion_tokens}, total_tokens={self.total_tokens})"

    def __repr__(self):
        return self.__str__()


class LLMOutput(BaseModel):
    """
    Base class for LLM outputs.
    """

    content: str = Field(..., description="Text of the LLM output.")
    usage: LLMUsage = Field(
        LLMUsage(), description="Usage information for the LLM output."
    )

    def __str__(self):
        return f"LLMOutput(content={self.content}, usage={self.usage})"

    def __repr__(self):
        return self.__str__()



from rag_colls.processors.readers.txt import TxtReader


def test_txt_reader():
    """
    Test the TXTReader class.
    """
    reader = TxtReader()

    documents = reader.load_data(file_path="samples/data/test.csv")

    assert len(documents) > 0, "No documents found in the TXT file."

    first_document = documents[0]
    assert hasattr(first_document, "document"), (
        "Document does not have document attribute."
    )
    assert hasattr(first_document, "metadata"), (
        "Document does not have metadata attribute."
    )

    assert "source" in first_document.metadata, "Metadata missing source"



from rag_colls.processors.readers.csv import CSVReader


def test_csv_reader():
    """
    Test the CSVReader class.
    """
    reader = CSVReader()

    documents = reader.load_data(file_path="samples/data/test.csv")

    assert len(documents) > 0, "No documents found in the CSV file."

    first_document = documents[0]
    assert hasattr(first_document, "document"), (
        "Document does not have document attribute."
    )
    assert hasattr(first_document, "metadata"), (
        "Document does not have metadata attribute."
    )

    # Test metadata
    metadata = first_document.metadata
    assert "file_path" in metadata, "Metadata missing file_path"
    assert "file_type" in metadata, "Metadata missing file_type"
    assert metadata["file_type"] == "csv", "Incorrect file type in metadata"
    assert "file_size" in metadata, "Metadata missing file_size"
    assert "num_rows" in metadata, "Metadata missing num_rows"
    assert "num_cols" in metadata, "Metadata missing num_cols"
    assert "columns" in metadata, "Metadata missing columns"
    assert "source" in metadata, "Metadata missing source"



from rag_colls.processors.readers.json import JSONReader


def test_json_reader():
    """
    Test the JSONReader class.
    """
    reader = JSONReader()

    documents = reader.load_data(file_path="samples/data/test.json")

    assert len(documents) > 0, "No documents found in the JSON file."

    first_document = documents[0]
    assert hasattr(first_document, "document"), (
        "Document does not have document attribute."
    )
    assert hasattr(first_document, "metadata"), (
        "Document does not have metadata attribute."
    )

    # Test metadata
    metadata = first_document.metadata
    assert "file_path" in metadata, "Metadata missing file_path"
    assert "file_type" in metadata, "Metadata missing file_type"
    assert metadata["file_type"] == "json", "Incorrect file type in metadata"
    assert "file_size" in metadata, "Metadata missing file_size"
    assert "encoding" in metadata, "Metadata missing encoding"
    assert "source" in metadata, "Metadata missing source"



from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed

from rag_colls.rags.base import BaseRAG
from rag_colls.core.base.chunkers.base import BaseChunker
from rag_colls.core.base.rerankers.base import BaseReranker
from rag_colls.core.base.llms.base import BaseCompletionLLM
from rag_colls.core.base.embeddings.base import BaseEmbedding
from rag_colls.core.base.database.bm25 import BaseBM25Retriever
from rag_colls.core.base.database.vector_database import BaseVectorDatabase

from rag_colls.prompts.q_a import Q_A_PROMPT
from rag_colls.types.core.document import Document
from rag_colls.types.llm import Message, LLMOutput
from rag_colls.core.settings import GlobalSettings
from rag_colls.core.utils import check_placeholders
from rag_colls.processors.file_processor import FileProcessor
from rag_colls.retrievers.bm25_retriever import BM25Retriever
from rag_colls.types.retriever import RetrieverIngestInput, RetrieverQueryType
from rag_colls.retrievers.vector_database_retriever import VectorDatabaseRetriever

from .utils import gen_contextual_chunk
from .prompt import CONTEXTUAL_PROMPT


class ContextualRAG(BaseRAG):
    """"""

    def __init__(
        self,
        *,
        vector_database: BaseVectorDatabase,
        bm25: BaseBM25Retriever,
        reranker: BaseReranker,
        chunker: BaseChunker,
        llm: BaseCompletionLLM | None = None,
        embed_model: BaseEmbedding | None = None,
        processor: FileProcessor | None = None,
        gen_contextual_prompt_template: str | None = None,
    ):
        self.vector_database = vector_database
        self.bm25 = bm25
        self.reranker = reranker
        self.chunker = chunker
        self.processor = processor or FileProcessor()

        self.embed_model = embed_model or GlobalSettings.embed_model
        self.llm = llm or GlobalSettings.completion_llm

        self.semantic_retriever = VectorDatabaseRetriever.from_vector_db(
            vector_db=vector_database, embed_model=self.embed_model
        )
        self.bm25_retriever = BM25Retriever.from_bm25(bm25=self.bm25)

        if gen_contextual_chunk:
            assert check_placeholders(
                template=gen_contextual_prompt_template,
                placeholders=["CHUNK_CONTENT", "WHOLE_DOCUMENT"],
            ), (
                f"Prompt template must contain the placeholders: {['CHUNK_CONTENT', 'WHOLE_DOCUMENT']}. Example: =======\n{CONTEXTUAL_PROMPT}"
            )

            self.gen_contextual_prompt_template = gen_contextual_prompt_template

        else:
            self.gen_contextual_prompt_template = CONTEXTUAL_PROMPT

    def _split_document(self, document: Document, **kwargs) -> list[Document]:
        """
        Split the document into chunks using the chunker.

        Args:
            document (Document): The document to be split.
            **kwargs: Additional arguments for the chunker.

        Returns:
            list[str]: A list of chunks.
        """
        return self.chunker.chunk(documents=[document], **kwargs)

    def _build_gen_context_input(
        self, chunks: list[Document], whole_document: Document
    ) -> list[tuple[Document, Document]]:
        """
        Build the input for generating contextual chunks.

        Args:
            chunks (list[Document]): The list of chunks.
            whole_document (Document): The whole document.

        Returns:
            list[tuple[Document, Document]]: A list of tuples containing the chunk and the whole document.
        """
        return [(chunk, whole_document) for chunk in chunks]

    def _ingest_db(
        self,
        file_or_folder_paths: list[str],
        batch_embedding: int = 100,
        num_workers: int = 4,
    ) -> None:
        """
        Ingest documents into the Contextual RAG database.

        Args:
            file_paths (list[str]): List of file paths to be ingested.
            batch_embedding (int): Batch size for embedding documents.
            num_workers (int): Number of workers for parallel processing.
        """
        documents = self.processor.load_data(file_or_folder_paths=file_or_folder_paths)

        chunks = []
        for doc in tqdm(
            documents, desc="Splitting documents into chunks ...", unit="doc"
        ):
            chunks.append(self._split_document(doc, show_progress=False))

        gen_contextual_inputs: list[tuple[Document, Document]] = []
        for document, chunk in zip(documents, chunks):
            gen_contextual_inputs.extend(self._build_gen_context_input(chunk, document))

        with ThreadPoolExecutor(max_workers=num_workers) as executor:
            futures = [
                executor.submit(
                    gen_contextual_chunk,
                    chunk,
                    whole_document,
                    self.llm,
                    self.gen_contextual_prompt_template,
                )
                for chunk, whole_document in gen_contextual_inputs
            ]

            contextual_chunks = []
            for future in tqdm(
                as_completed(futures),
                total=len(futures),
                desc=f"Generating contextual chunks (num_workers={num_workers})",
            ):
                contextual_chunks.append(future.result())

        embeddings = self.embed_model.get_batch_document_embedding(
            documents=contextual_chunks, batch_size=batch_embedding
        )

        embeded_chunks = [
            RetrieverIngestInput(
                id=doc.id,
                document=doc.document,
                embedding=e.embedding,
                metadata=doc.metadata,
            )
            for doc, e in zip(contextual_chunks, embeddings)
        ]

        self.vector_database.add_documents(
            documents=embeded_chunks,
        )

        self.bm25.add_documents(
            documents=embeded_chunks,
        )

    def _search(self, query: RetrieverQueryType, top_k: int = 5) -> LLMOutput:
        """
        Search with Contextual RAG.

        Args:
            query (RetrieverQueryType): The query to search for.
            top_k (int): The number of top results to retrieve.

        Returns:
            LLMOutput: The response from the LLM.
        """
        with ThreadPoolExecutor(max_workers=2) as executor:
            future_semantic = executor.submit(
                self.semantic_retriever.retrieve, query=query, top_k=top_k
            )
            future_bm25 = executor.submit(
                self.bm25_retriever.retrieve, query=query, top_k=top_k
            )

            semantic_results = future_semantic.result()
            bm25_results = future_bm25.result()

        reranked_results = self.reranker.rerank(
            query=query,
            results=[semantic_results, bm25_results],
            top_k=top_k,
        )

        contexts = ""
        for result in reranked_results:
            contexts += f"{result.document} \n ============ \n"

        messages = [
            Message(
                role="user", content=Q_A_PROMPT.format(context=contexts, question=query)
            )
        ]

        response = self.llm.complete(messages=messages)

        return response



from rag_colls.llms.litellm_llm import LiteLLM
from rag_colls.databases.bm25.bm25s import BM25s
from rag_colls.embeddings.openai_embedding import OpenAIEmbedding
from rag_colls.rerankers.weighted_reranker import WeightedReranker
from rag_colls.processors.chunkers.semantic_chunker import SemanticChunker
from rag_colls.databases.vector_databases.chromadb import ChromaVectorDatabase

from rag_colls.rags.contextual_rag import ContextualRAG, CONTEXTUAL_PROMPT

rag = ContextualRAG(
    vector_database=ChromaVectorDatabase(
        persistent_directory="./chroma_db", collection_name="test"
    ),
    bm25=BM25s(
        save_dir="./bm25s",
    ),
    reranker=WeightedReranker(weights=[0.8, 0.2]),  # [semantic_weight, bm25_weight]
    chunker=SemanticChunker(embed_model_name="text-embedding-ada-002"),
    llm=LiteLLM(model_name="openai/gpt-4o-mini"),
    embed_model=OpenAIEmbedding(model_name="text-embedding-ada-002"),
    gen_contextual_prompt_template=CONTEXTUAL_PROMPT,
)

response = rag.search(query="What is Chain Buddy ?", top_k=5)

print(response.content)
print("===========")
print(response.usage)



# REF: https://github.com/run-llama/llama_index/blob/main/llama-index-integrations/readers/llama-index-readers-file/llama_index/readers/file/tabular/base.py
from typing import Any, List, Optional, Dict
from pathlib import Path

from rag_colls.types.core.document import Document
from rag_colls.core.base.readers.base import BaseReader


class CSVReader(BaseReader):
    """CSV parser.

    Args:
        concat_rows (bool): whether to concatenate all rows into one document.
            If set to False, a Document will be created for each row.
            True by default.

    """

    def __init__(self, *args: Any, concat_rows: bool = True, **kwargs: Any) -> None:
        """Init params."""
        super().__init__(*args, **kwargs)
        self._concat_rows = concat_rows

    def _load_data(
        self,
        file: Path,
        should_split: bool = True,
        extra_info: Optional[Dict] = None,
        encoding: str = "utf-8",
    ) -> List[Document]:
        """Parse file.

        Returns:
            Union[str, List[str]]: a string or a List of strings.

        """
        try:
            import csv
        except ImportError:
            raise ImportError(
                "csv module is not installed. Please install it with 'pip install csv'."
            )

        if isinstance(file, str):
            file = Path(file)

        text_list = []
        with open(file) as fp:
            csv_reader = csv.reader(fp)
            for row in csv_reader:
                text_list.append(", ".join(row))

        metadata = {
            "filename": file.name,
            "extension": file.suffix,
            "source": f"{file.name}",
            "should_split": should_split,
            "file_size": file.stat().st_size,
            "file_path": str(file),
            "file_type": "csv",
            "encoding": encoding,
            "num_rows": len(text_list),
            "num_cols": len(text_list[0].split(",")),
            "columns": text_list[0].split(","),
        }
        if extra_info:
            metadata = {**metadata, **extra_info}
        if self._concat_rows:
            return [
                Document(
                    document="\n".join(text_list).encode(encoding), metadata=metadata
                )
            ]
        else:
            return [
                Document(document=text.encode(encoding), metadata=metadata)
                for text in text_list
            ]



from rag_colls.processors.file_processor import FileProcessor
from pathlib import Path


def test_file_processor():
    file_processor = FileProcessor()
    file_paths = [str(file) for file in Path("samples/data").glob("*")]
    documents = file_processor.load_data(
        file_or_folder_paths=file_paths,
        should_splits=[True] * len(file_paths),
        extra_infos=[None] * len(file_paths),
    )
    assert len(documents) > 0, "No documents found in the data directory."
    print("len(file_paths):", len(file_paths))
    print("len(documents):", len(documents))
    for i in range(len(documents)):
        print("documents[{}].metadata:".format(i), documents[i].metadata)


def test_file_processor_dir():
    file_processor = FileProcessor()
    file_paths = ["samples"]
    documents = file_processor.load_data(
        file_or_folder_paths=file_paths,
    )
    assert len(documents) > 0, "No documents found in the data directory."
    print("len(documents):", len(documents))
    for i in range(len(documents)):
        print("documents[{}].metadata:".format(i), documents[i].metadata)



from rag_colls.core.base.rerankers.base import BaseReranker

from rag_colls.types.reranker import RerankerResult
from rag_colls.types.retriever import RetrieverQueryType, RetrieverResult


class WeightedReranker(BaseReranker):
    """
    Weighted Reranker that assigns weights to the results based on their relevance score.
    """

    def __init__(self, weights: list[float]):
        """
        Initialize the WeightedReranker with weights.

        Args:
            weights (list[float]): List of weights for each result.
        """
        self.weights = weights

    def _rerank(
        self,
        query: RetrieverQueryType,
        results: list[list[RetrieverResult]],
        top_k: int = 10,
    ) -> list[RerankerResult]:
        """
        Rerank the results based on the query and weights.

        Args:
            query (RetrieverQueryType): The query to rerank the results for.
            results (list[list[RetrieverResult]]): The results to rerank.
            top_k (int): The number of top results to return.

        Returns:
            list[RerankerResult]: The reranked results.
        """
        reranked_results: list[RerankerResult] = []

        assert len(results) == len(self.weights), (
            "Number of results must match number of weights."
        )

        # Make sure weights are normalized to sum to 1
        total_weight = sum(self.weights)

        assert total_weight > 0, "Weights must sum to a positive value."

        if total_weight != 1:
            self.weights = [w / total_weight for w in self.weights]

        for i, result_set in enumerate(results):
            weighted_results = [
                RerankerResult(
                    id=result.id,
                    document=result.document,
                    score=result.score * self.weights[i],
                    metadata=result.metadata,
                )
                for result in result_set
            ]

            reranked_results.extend(weighted_results)

        reranked_results.sort(key=lambda x: x.score, reverse=True)

        return reranked_results[:top_k]



from loguru import logger
from litellm import completion, acompletion

from rag_colls.core.constants import DEFAULT_OPENAI_MODEL
from rag_colls.core.base.llms.base import BaseCompletionLLM
from rag_colls.types.llm import Message, LLMOutput, LLMUsage


class LiteLLM(BaseCompletionLLM):
    """
    A lightweight wrapper for the litellm library.

    litellm provide many models from openai, anthropic, google, etc..
    """

    def __init__(self, model_name: str | None = None):
        """
        Initialize the LiteLLM class.

        Args:
            model_name (str): The name of the model to use.
        """
        self.model_name = model_name or DEFAULT_OPENAI_MODEL
        logger.info(f"Using LiteLLM with model: {self.model_name}")

    def __str__(self):
        return f"LiteLLM(model_name={self.model_name})"

    def __repr__(self):
        return self.__str__()

    def _complete(self, messages: list[Message], **kwargs) -> LLMOutput:
        formatted_messages = [
            {"role": msg.role, "content": msg.content} for msg in messages
        ]

        response = completion(
            model=self.model_name, messages=formatted_messages, **kwargs
        )
        return LLMOutput(
            content=response.choices[0].message.content,
            usage=LLMUsage(
                prompt_tokens=response.usage.prompt_tokens,
                completion_tokens=response.usage.completion_tokens,
                total_tokens=response.usage.total_tokens,
            ),
        )

    async def _acomplete(self, messages: list[Message], **kwargs) -> LLMOutput:
        formatted_messages = [
            {"role": msg.role, "content": msg.content} for msg in messages
        ]
        response = await acompletion(
            model=self.model_name, messages=formatted_messages, **kwargs
        )

        return LLMOutput(
            content=response.choices[0].message.content,
            usage=LLMUsage(
                prompt_tokens=response.usage.prompt_tokens,
                completion_tokens=response.usage.completion_tokens,
                total_tokens=response.usage.total_tokens,
            ),
        )



from pathlib import Path
from typing import Optional

from rag_colls.types.core.document import Document
from rag_colls.core.base.readers.base import BaseReader


class HTMLReader(BaseReader):
    """Reader HTML usimg html2text

    Reader behavior:
        - HTML is read with html2text.
        - All of the texts will be split by `page_break_pattern`
        - Each page is extracted as a Document
        - The output is a list of Documents

    Args:
        page_break_pattern (str): Pattern to split the HTML into pages
    """

    def __init__(self, page_break_pattern: Optional[str] = None, *args, **kwargs):
        self._page_break_pattern: Optional[str] = page_break_pattern
        super().__init__()

    def _load_data(
        self,
        file_path: Path | str,
        should_split: bool = True,
        extra_info: Optional[dict] = None,
        encoding: str = "utf-8",
    ) -> list[Document]:
        """Load data using Html reader

        Args:
            file_path: path to HTML file
            should_split: whether to split the HTML into pages
            extra_info: extra information passed to this reader during extracting data

        Returns:
            list[Document]: list of documents extracted from the HTML file
        """
        try:
            import html2text  # noqa
        except ImportError:
            raise ImportError(
                "html2text is not installed. "
                "Please install it using `pip install html2text`"
            )

        file_path = Path(file_path).resolve()

        with file_path.open("r") as f:
            html_text = "".join([line[:-1] for line in f.readlines()])

        # read HTML
        all_text = html2text.html2text(html_text)
        pages = (
            all_text.split(self._page_break_pattern)
            if self._page_break_pattern
            else [all_text]
        )

        extra_info = extra_info or {
            "file_path": str(file_path),
            "should_split": should_split,
            "file_size": file_path.stat().st_size,
            "file_type": "html",
            "encoding": encoding,
        }

        documents = []
        for page_id, page in enumerate(pages):
            page_extra_info = extra_info.copy()
            page_extra_info["source"] = f"{file_path.name}: Page {page_id + 1}"
            documents.append(
                Document(
                    document=page.strip().encode(encoding), metadata=page_extra_info
                )
            )

        return documents



from pathlib import Path

from rag_colls.types.core.document import Document
from rag_colls.core.base.readers.base import BaseReader


class PyMuPDFReader(BaseReader):
    def _load_data(
        self,
        file_path: str | Path,
        should_split: bool = True,
        extra_info: dict | None = None,
        encoding: str = "utf-8",
    ) -> list[Document]:
        try:
            import fitz
        except ImportError:
            raise ImportError(
                "PyMuPDF is not installed. Please install it with 'pip install PyMuPDF'."
            )

        docs = fitz.open(file_path)
        file_name = Path(file_path).name

        if not extra_info:
            extra_info = {}

        extra_info["file_path"] = file_path
        extra_info["should_split"] = should_split

        documents = []
        for doc in docs:
            documents.append(
                Document(
                    document=doc.get_text().encode(encoding),
                    metadata=dict(
                        source=f"{file_name}: Page {doc.number + 1}",
                        **extra_info,
                    ),
                )
            )

        docs.close()
        return documents



import asyncio
import chromadb
from loguru import logger

from rag_colls.types.retriever import RetrieverIngestInput, RetrieverResult
from rag_colls.core.base.database.vector_database import BaseVectorDatabase


class ChromaVectorDatabase(BaseVectorDatabase):
    def __init__(self, persistent_directory: str, collection_name: str):
        """
        Initialize the Chroma vector database.

        Args:
            persistent_directory (str): Directory to persist the database.
            collection_name (str): Name of the collection to create.
        """

        self.client = chromadb.PersistentClient(
            path=persistent_directory,
        )
        self._test_connection()

        self.collection_name = collection_name

        self._create_collection(collection_name)

        logger.success("ChromaVectorDatabase initialized successfully !!!")

    def _test_connection(self):
        """
        Test the connection to the Chroma vector database.
        """
        self.client.heartbeat()

    def _check_collection_exists(self, collection_name):
        """
        Check if a collection exists in the Chroma vector database.

        Args:
            collection_name (str): Name of the collection to check.
        """
        try:
            self.client.get_collection(name=collection_name)
        except Exception:
            return False
        return True

    def _create_collection(self, collection_name, **kwargs):
        """
        Create a collection in the Chroma vector database.

        Args:
            collection_name (str): Name of the collection to create.
            **kwargs: Additional arguments for the collection.
        """
        if self._check_collection_exists(collection_name):
            return

        self.client.create_collection(
            name=collection_name,
            metadata={"hnsw:space": "cosine"},
        )
        logger.success(f"Collection {collection_name} created successfully !!!")

    def _add_documents(self, documents: list[RetrieverIngestInput], **kwargs):
        """
        Add documents to the Chroma vector database.

        Args:
            documents (list[str]): List of documents to add.
            **kwargs: Additional arguments for the add operation.
        """
        if not self._check_collection_exists(self.collection_name):
            logger.error(f"Collection {self.collection_name} does not exist.")
            return

        collection = self.client.get_collection(self.collection_name)

        metadatas = [{"document": doc.document, **doc.metadata} for doc in documents]
        embeddings = [doc.embedding for doc in documents]

        collection.add(
            embeddings=embeddings,
            metadatas=metadatas,
            ids=[doc.id for doc in documents],
        )

        logger.success(f"Added: {len(documents)} documents.")

    def _delete_documents(self, document_ids: list[str], **kwargs):
        """
        Delete documents from the Chroma vector database.

        Args:
            document_ids (list[str]): List of document IDs to delete.
            **kwargs: Additional arguments for the delete operation.
        """
        if not self._check_collection_exists(self.collection_name):
            logger.error(f"Collection {self.collection_name} does not exist.")
            return

        collection = self.client.get_collection(self.collection_name)
        collection.delete(ids=document_ids)

        logger.success(f"Deleted: {len(document_ids)} documents.")

    def _search(
        self, query_embedding: list[float], top_k: int = 5, **kwargs
    ) -> list[RetrieverResult]:
        """
        Search for documents in the Chroma vector database.

        Args:
            query_embedding (list[float]): The embedding of the query.
            top_k (int): Number of top results to return.
            **kwargs: Additional arguments for the search operation.

        Returns:
            list[RetrieverResult]: List of retrieved documents with their metadata.
        """
        if not self._check_collection_exists(self.collection_name):
            logger.error(f"Collection {self.collection_name} does not exist.")
            return []

        collection = self.client.get_collection(self.collection_name)
        results = collection.query(query_embeddings=query_embedding, n_results=top_k)

        ids = results["ids"][0]
        metadatas = results["metadatas"][0]

        # We use cosine_similarity to have larger score for more similar documents
        scores = [1 - s for s in results["distances"][0]]

        documents = [
            RetrieverResult(
                id=doc_id,
                score=score,
                document=metadata["document"],
                metadata=metadata,
            )
            for doc_id, metadata, score in zip(ids, metadatas, scores)
        ]

        return documents

    async def _asearch(
        self, query_embedding: list[float], top_k: int = 5, **kwargs
    ) -> list[RetrieverResult]:
        """
        Asynchronous search for documents in the Chroma vector database.

        Args:
            query_embedding (list[float]): The embedding of the query.
            top_k (int): Number of top results to return.
            **kwargs: Additional arguments for the search operation.

        Returns:
            list[RetrieverResult]: List of retrieved documents with their metadata.
        """
        return await asyncio.to_thread(
            self._search,
            query_embedding=query_embedding,
            top_k=top_k,
            **kwargs,
        )



# Configuration file for the Sphinx documentation builder.

# -- Project information
import os
import sys

project = "rag-colls"
copyright = "2025, hienhayho"
author = "hienhayho"

release = "0.2.0.1"
version = "0.2.0.1"

sys.path.append(os.path.abspath("../.."))

# -- General configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "myst_parser",
    "sphinx_copybutton",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

# -- Options for HTML output

html_baseurl = os.environ.get("READTHEDOCS_CANONICAL_URL", "/")
# html_theme = "furo"
html_theme = "sphinx_book_theme"

html_logo = "_static/Final_logo.png"
html_theme_options = {
    "path_to_docs": "docs/source",
    "repository_url": "https://github.com/hienhayho/rag-colls",
    "use_repository_button": True,
    "use_edit_page_button": True,
}

html_static_path = ["_static"]

# -- Options for EPUB output
epub_show_urls = "footnote"
html_js_files = [
    ("readthedocs.js", {"defer": "defer"}),
]



from abc import ABC, abstractmethod
from rag_colls.types.embedding import Embedding
from rag_colls.types.core.document import Document
from tenacity import retry, wait_random_exponential, stop_after_attempt


class BaseEmbedding(ABC):
    @abstractmethod
    def _get_query_embedding(self, query: str) -> Embedding:
        """
        Returns the embedding of the query.

        Args:
            query (str): The query to be embedded.

        Returns:
            Embedding: The embedding object of the query.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @abstractmethod
    def _get_document_embedding(self, document: Document) -> Embedding:
        """
        Returns the embedding of the document.

        Args:
            document (Document): The document to be embedded.

        Returns:
            list: The embedding of the document.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @abstractmethod
    def _get_batch_query_embedding(
        self, queries: list[str], **kwargs
    ) -> list[Embedding]:
        """
        Returns the embeddings of the queries.

        Args:
            queries (list[str]): The list of queries to be embedded.
            **kwargs: Additional keyword arguments for the embedding function.

        Returns:
            list[Embedding]: The list of embedding objects of the queries.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @abstractmethod
    def _get_batch_document_embedding(
        self, documents: list[Document], **kwargs
    ) -> list[Embedding]:
        """
        Returns the embeddings of the documents.

        Args:
            documents (list[Document]): The list of documents to be embedded.

        Returns:
            list[Embedding]: The list of embedding objects of the documents.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(6))
    def get_query_embedding(self, query: str) -> Embedding:
        """
        Returns the embedding of the query.

        Args:
            query (str): The query to be embedded.

        Returns:
            Embedding: The embedding object of the query.
        """
        return self._get_query_embedding(query)

    @retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(6))
    def get_document_embedding(self, document: Document) -> Embedding:
        """
        Returns the embedding of the document.

        Args:
            document (Document): The document to be embedded.

        Returns:
            Embedding: The embedding object of the document.
        """
        return self._get_document_embedding(document)

    @retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(6))
    def get_batch_query_embedding(
        self, queries: list[str], **kwargs
    ) -> list[Embedding]:
        """
        Returns the embeddings of the queries.

        Args:
            queries (list[str]): The list of queries to be embedded.
            **kwargs: Additional keyword arguments for the embedding function.

        Returns:
            list[Embedding]: The list of embedding objects of the queries.
        """
        return self._get_batch_query_embedding(queries, **kwargs)

    @retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(6))
    def get_batch_document_embedding(
        self, documents: list[Document], **kwargs
    ) -> list[Embedding]:
        """
        Returns the embeddings of the documents.

        Args:
            documents (list[Document]): The list of documents to be embedded.
            **kwargs: Additional keyword arguments for the embedding function.

        Returns:
            list[Embedding]: The list of embedding objects of the documents.
        """
        return self._get_batch_document_embedding(documents, **kwargs)



from rag_colls.core.base.retrievers.base import BaseRetriever
from rag_colls.core.base.database.bm25 import BaseBM25Retriever

from rag_colls.types.retriever import RetrieverQueryType, RetrieverResult


class BM25Retriever(BaseRetriever):
    """
    A retriever that uses BM25 to retrieve documents based on a query.
    """

    def __init__(self, bm25: BaseBM25Retriever):
        """
        Initialize the BM25Retriever class.

        Args:
            bm25: The BM25 retriever to use for retrieval.
            embed_model: The embedding model to use for generating embeddings.
        """
        self.bm25 = bm25

    @classmethod
    def from_bm25(cls, bm25: BaseBM25Retriever) -> "BM25Retriever":
        """
        Create an instance of BM25Retriever from a BM25 retriever.

        Args:
            bm25: The BM25 retriever to use for retrieval.

        Returns:
            BM25Retriever: An instance of the retriever.
        """
        return cls(bm25=bm25)

    def _retrieve(self, query: RetrieverQueryType, **kwargs) -> list[RetrieverResult]:
        """
        Retrieve documents based on the query.

        Args:
            query: The query to search for.
            **kwargs: Additional keyword arguments for the retrieval process.

        Returns:
            list[RetrieverResult]: A list of retrieved documents.
        """
        assert isinstance(query, str), "Query must be a string."

        return self.bm25.search(query=query, **kwargs)



from abc import ABC, abstractmethod
from rag_colls.types.llm import LLMOutput


class BaseRAG(ABC):
    @abstractmethod
    def _ingest_db(self, file_or_folder_paths: list[str], **kwargs):
        """
        Ingest documents process

        Args:
            file_or_folder_paths (list[str]): List of file paths or folders to be ingested.
            **kwargs: Additional keyword arguments for the ingestion process.
        """
        raise NotImplementedError("Ingesting documents process is not implemented.")

    @abstractmethod
    def _search(self, query: str, **kwargs) -> LLMOutput:
        """
        Search for the most relevant documents based on the query.

        Args:
            query (str): The query to search for.
            **kwargs: Additional keyword arguments for the search operation.

        Returns:
            LLMOutput: The response from the LLM.
        """
        raise NotImplementedError("Searching documents process is not implemented.")

    def ingest_db(self, file_or_folder_paths: list[str], **kwargs):
        """
        Ingest documents into the vector database.

        Args:
            file_or_folder_paths (list[str]): List of file paths or folders to be ingested.
            **kwargs: Additional keyword arguments for the ingestion process.
        """
        return self._ingest_db(file_or_folder_paths=file_or_folder_paths, **kwargs)

    def search(self, query: str, **kwargs) -> LLMOutput:
        """
        Search for the most relevant documents based on the query.

        Args:
            query (str): The query to search for.
            **kwargs: Additional keyword arguments for the search operation.

        Returns:
            LLMOutput: The response from the LLM.
        """
        return self._search(query, **kwargs)



from rag_colls.rags.base import BaseRAG
from rag_colls.core.base.chunkers.base import BaseChunker
from rag_colls.core.base.embeddings.base import BaseEmbedding
from rag_colls.core.base.llms.base import BaseCompletionLLM
from rag_colls.core.base.database.vector_database import BaseVectorDatabase


from rag_colls.prompts.q_a import Q_A_PROMPT
from rag_colls.types.llm import Message, LLMOutput
from rag_colls.core.settings import GlobalSettings
from rag_colls.types.retriever import RetrieverIngestInput
from rag_colls.processors.file_processor import FileProcessor
from rag_colls.retrievers.vector_database_retriever import VectorDatabaseRetriever


class BasicRAG(BaseRAG):
    """
    Wrapper class for a basic RAG (Retrieval-Augmented Generation) system. This class integrates a vector database, chunker, LLM, and embedding model to perform semantic search.

    This is helpful for anyone begin starting with RAG and understanding how to use it.
    """

    def __init__(
        self,
        *,
        vector_database: BaseVectorDatabase,
        chunker: BaseChunker,
        llm: BaseCompletionLLM | None = None,
        embed_model: BaseEmbedding | None = None,
        processor: FileProcessor | None = None,
    ):
        """
        Initialize the BasicRAG class.

        Args:
            vector_database (BaseVectorDatabase): The vector database to use for storing and retrieving documents.
            chunker (BaseChunker): The chunker to use for splitting documents into smaller chunks.
            llm (BaseCompletionLLM, optional): The LLM to use for generating responses. Defaults to `None`.
            embed_model (BaseEmbedding, optional): The embedding model to use for generating embeddings. Defaults to `None`.
            processor: (FileProcessor, optional): The processor to use for loading and processing documents. Defaults to `None`.
        """
        self.vector_database = vector_database
        self.chunker = chunker
        self.processor = processor or FileProcessor()
        self.embed_model = embed_model or GlobalSettings.embed_model
        self.llm = llm or GlobalSettings.completion_llm

        self.retriever = VectorDatabaseRetriever.from_vector_db(
            vector_db=vector_database, embed_model=self.embed_model
        )

    def _ingest_db(self, file_or_folder_paths: list[str], batch_embedding: int = 100):
        """
        Ingest documents into the vector database.

        Args:
            file_paths (list[str]): List of file paths to be ingested.
            batch_embedding (int): Batch size for embedding documents.
        """
        documents = self.processor.load_data(file_or_folder_paths=file_or_folder_paths)

        chunks = self.chunker.chunk(documents=documents)

        embeddings = self.embed_model.get_batch_document_embedding(
            documents=chunks, batch_size=batch_embedding
        )

        embeded_chunks = [
            RetrieverIngestInput(
                id=doc.id,
                document=doc.document,
                embedding=e.embedding,
                metadata=doc.metadata,
            )
            for doc, e in zip(chunks, embeddings)
        ]

        self.vector_database.add_documents(
            documents=embeded_chunks,
        )

    def _search(self, query: str, top_k: int = 5, **kwargs) -> LLMOutput:
        """
        Search for the most relevant documents based on the query.

        Args:
            query (str): The query to search for.
            top_k (int): The number of top results to retrieve.
            **kwargs: Additional keyword arguments for the search operation.

        Returns:
            LLMOutput: The response from the LLM.
        """

        results = self.retriever.retrieve(
            query=query,
            k=top_k,
        )

        contexts = ""

        for result in results:
            contexts += f"{result.document}\n ============ \n"

        messages = [
            Message(
                role="user", content=Q_A_PROMPT.format(context=contexts, question=query)
            )
        ]

        response = self.llm.complete(messages=messages)

        return response



from typing import Union, Optional
from pydantic import BaseModel, Field

EmbeddingType = list[float]
RetrieverQueryType = Union[str, EmbeddingType]


class RetrieverIngestInput(BaseModel):
    id: str = Field(
        ..., description="The unique identifier for the document to be ingested."
    )
    document: str = Field(
        ..., description="The content of the document to be ingested."
    )
    embedding: Optional[list[float]] = Field(
        description="The embedding vector of the document to be ingested. Could be None.",
    )
    metadata: dict = Field(
        default_factory=dict,
        description="Metadata associated with the document to be ingested.",
    )

    def __str__(self):
        return f"IngestInput(id={self.id}, dimensions={len(self.embedding)}, metadata={self.metadata})"

    def __repr__(self):
        return self.__str__()


class RetrieverResult(BaseModel):
    """
    A class representing the result of a retrieval operation.
    """

    id: str = Field(..., description="The unique identifier of the retrieved document.")
    score: float = Field(
        ..., description="The score or relevance of the retrieved document."
    )
    document: str = Field(..., description="The content of the retrieved document.")
    metadata: dict = Field(
        default_factory=dict,
        description="Metadata associated with the retrieved document.",
    )

    def __str__(self):
        return f"RetrieveResult(id={self.id}, score={self.score}, metadata={self.metadata})"

    def __repr__(self):
        return self.__str__()



from rag_colls.core.base.embeddings.base import BaseEmbedding
from rag_colls.core.base.retrievers.base import BaseRetriever
from rag_colls.core.base.database.vector_database import BaseVectorDatabase

from rag_colls.core.settings import GlobalSettings
from rag_colls.types.retriever import RetrieverQueryType, RetrieverResult


class VectorDatabaseRetriever(BaseRetriever):
    """
    A retriever that uses vector databases to retrieve documents based on a query.
    """

    def __init__(
        self, vector_db: BaseVectorDatabase, embed_model: BaseEmbedding | None = None
    ):
        """
        Initialize the VectorDatabaseRetriever class.

        Args:
            vector_db: The vector database to use for retrieval.
            embed_model: The embedding model to use for generating embeddings.
        """
        self.vector_db = vector_db
        self.embed_model = embed_model or GlobalSettings.embed_model

    @classmethod
    def from_vector_db(
        cls, vector_db: BaseVectorDatabase, embed_model: BaseEmbedding | None = None
    ) -> "VectorDatabaseRetriever":
        """
        Create an instance of VectorDatabaseRetriever from a vector database.

        Args:
            vector_db: The vector database to use for retrieval.

        Returns:
            VectorDatabaseRetriever: An instance of the retriever.
        """
        return cls(vector_db=vector_db, embed_model=embed_model)

    def _retrieve(self, query: RetrieverQueryType, **kwargs) -> list[RetrieverResult]:
        """
        Retrieve documents based on the query.

        Args:
            query: The query to search for.
            **kwargs: Additional keyword arguments for the retrieval process.

        Returns:
            list[RetrieverResult]: A list of retrieved documents.
        """
        query_embedding = query
        if isinstance(query, str):
            query_embedding = self.embed_model.get_query_embedding(query).embedding

        return self.vector_db.search(query_embedding=query_embedding, **kwargs)



from rag_colls.types.llm import Message
from rag_colls.types.core.document import Document
from rag_colls.core.base.llms.base import BaseCompletionLLM

from .prompt import CONTEXTUAL_PROMPT


def gen_contextual_chunk(
    chunk: Document,
    whole_document: Document,
    llm: BaseCompletionLLM,
    contextual_prompt_template: str = CONTEXTUAL_PROMPT,
) -> Document:
    """
    Generate a contextual document using the provided LLM.

    Args:
        chunk (Document): The chunk of text to be contextualized.
        whole_document (Document): The whole document containing the context.
        llm (BaseCompletionLLM): The language model used for generation.
        contextual_prompt_template (str): The template for the contextual prompt.

    Returns:
        Document: The generated contextual document.
    """
    messages = [
        Message(
            role="user",
            content=contextual_prompt_template.format(
                CHUNK_CONTENT=chunk.document, WHOLE_DOCUMENT=whole_document.document
            ),
        )
    ]

    response = llm.complete(messages)

    return Document(
        id=chunk.id,
        document=response.content + "\n\n" + chunk.document,
        metadata=chunk.metadata,
    )



from abc import ABC, abstractmethod
from rag_colls.types.retriever import RetrieverQueryType, RetrieverResult


class BaseRetriever(ABC):
    """
    Base class for all retrievers.
    """

    @abstractmethod
    def _retrieve(self, query: RetrieverQueryType, **kwargs) -> list[RetrieverResult]:
        """
        Retrieve documents based on the query.

        Args:
            query (RetrieverQueryType): The query to search for.
            **kwargs: Additional keyword arguments for the retrieval process.

        Returns:
            list[RetrieverResult]: A list of retrieved documents.
        """
        raise NotImplementedError("Retrieving documents process is not implemented.")

    def retrieve(self, query: RetrieverQueryType, **kwargs) -> list[RetrieverResult]:
        """
        Retrieve documents based on the query.

        Args:
            query (RetrieverQueryType): The query to search for.
            **kwargs: Additional keyword arguments for the retrieval process.

        Returns:
            list[RetrieverResult]: A list of retrieved documents.
        """
        return self._retrieve(query, **kwargs)



import asyncio
from loguru import logger
from llama_index.embeddings.openai import OpenAIEmbedding
from llama_index.core.embeddings.mock_embed_model import MockEmbedding

from llama_index.core import Document as LlamaIndexDocument
from llama_index.core.node_parser import SemanticSplitterNodeParser

from rag_colls.types.core.document import Document
from rag_colls.core.base.chunkers.base import BaseChunker
from rag_colls.core.constants import (
    HF_EMBEDDING_MODELS,
    OPENAI_EMBEDDING_MODELS,
    DEFAULT_OPENAI_EMBEDDING_MODEL,
)


class SemanticChunker(BaseChunker):
    """
    Semantic chunker that chunks documents based on semantic similarity.
    """

    def __init__(
        self,
        embed_model_name: str | None = None,
        buffer_size: int = 1,
        breakpoint_percentile_threshold: int = 95,
        mocking: bool = False,
    ):
        if not embed_model_name:
            embed_model_name = DEFAULT_OPENAI_EMBEDDING_MODEL

        assert embed_model_name in OPENAI_EMBEDDING_MODELS + HF_EMBEDDING_MODELS, (
            f"Model {embed_model_name} is not supported. Please use openai or flag embedding models."
        )

        self.embed_model_name = embed_model_name
        self.buffer_size = buffer_size
        self.breakpoint_percentile_threshold = breakpoint_percentile_threshold

        if mocking:
            # NOTE: Use in ci/cd testing
            self.embed_model = MockEmbedding(embed_dim=512)

        else:
            if embed_model_name in OPENAI_EMBEDDING_MODELS:
                self.embed_model = OpenAIEmbedding(model=embed_model_name)
            else:
                from llama_index.embeddings.huggingface import HuggingFaceEmbedding

                self.embed_model = HuggingFaceEmbedding(model_name=embed_model_name)

        self.node_parser = SemanticSplitterNodeParser(
            buffer_size=self.buffer_size,
            breakpoint_percentile_threshold=self.breakpoint_percentile_threshold,
            embed_model=self.embed_model,
        )

        logger.success(
            f"SemanticChunker initialized with: {embed_model_name}",
        )

    def __str__(self):
        return f"SemanticChunker(embed_model_name={self.embed_model_name}, buffer_size={self.buffer_size}, breakpoint_percentile_threshold={self.breakpoint_percentile_threshold}), mocking={self.mocking})"

    def _chunk(self, documents: list[Document], show_progress: bool = True, **kwargs):
        """
        Chunk the documents based on semantic similarity.

        Args:
            documents (list[Document]): List of documents to be chunked.
            **kwargs: Additional keyword arguments for the chunking function.

        Returns:
            list[Document]: List of chunked documents.
        """
        preprocessed_documents = [
            LlamaIndexDocument(
                doc_id=doc.id,
                text=doc.document,
                metadata=doc.metadata,
            )
            for doc in documents
        ]

        nodes = self.node_parser.get_nodes_from_documents(
            documents=preprocessed_documents, show_progress=show_progress
        )

        return [Document(document=node.text, metadata=node.metadata) for node in nodes]

    async def _achunk(
        self, documents: list[Document], show_progress: bool = False, **kwargs
    ):
        """
        Asynchronously chunk the documents based on semantic similarity.

        Args:
            documents (list[Document]): List of documents to be chunked.
            **kwargs: Additional keyword arguments for the chunking function.

        Returns:
            list[Document]: List of chunked documents.
        """
        # For now, we will just call the synchronous method
        return await asyncio.to_thread(self._chunk, documents, show_progress, **kwargs)



import warnings

from typing import Any
from pathlib import Path
from loguru import logger
from multiprocessing import Pool
from rag_colls.types.core.document import Document
from rag_colls.core.base.readers.base import BaseReader


def process_file_worker(
    args: tuple[str | Path, bool, dict[str, Any], dict[str, BaseReader]],
) -> list[Document]:
    """
    Worker function for multiprocessing.

    Args:
        args (tuple): (file_path, should_split, extra_info, processors)

    Returns:
        list[Document]: List of processed documents.
    """
    file_path, should_split, extra_info, processors = args
    ext = Path(file_path).suffix.lower()

    if ext not in processors:
        raise ValueError(f"No processor found for file type: {ext}")

    reader = processors[ext]
    return reader.load_data(
        file_path=file_path, should_split=should_split, extra_info=extra_info
    )


async def process_file_worker_async(
    args: tuple[str | Path, bool, dict[str, Any], dict[str, BaseReader]],
) -> list[Document]:
    """
    Asynchronous worker function for processing files.
    Args:
        args (tuple): (file_path, should_split, extra_info, processors)
    Returns:
        list[Document]: List of processed documents.
    """
    file_path, should_split, extra_info, processors = args
    ext = Path(file_path).suffix.lower()

    if ext not in processors:
        raise ValueError(f"No processor found for file type: {ext}")

    reader = processors[ext]
    return await reader.aload_data(
        file_path=file_path, should_split=should_split, extra_info=extra_info
    )


class FileProcessor:
    def __init__(
        self,
        processors: dict[str, BaseReader] | None = None,
        merge_with_default_processors: bool = False,
    ):
        """
        Initialize the FileProcessor with a dictionary of file type processors.
        """
        self.processors = processors or {}

        if not processors:
            logger.info("No processors provided. Using default processors ...")
            merge_with_default_processors = True

        if merge_with_default_processors:
            default_processors = self._get_default_processors()
            for ext, processor in default_processors.items():
                if ext not in self.processors:
                    self.processors[ext] = processor

    def __str__(self):
        return "FileProcessor"

    def _get_default_processors(self) -> dict[str, BaseReader]:
        """
        Initialize default file processors.

        Returns:
            dict[str, BaseReader]: A dictionary of default file processors.
        """
        logger.info("Initializing default file processors ...")
        from .readers.pdf import PyMuPDFReader
        from .readers.csv import CSVReader
        from .readers.docx import DocxReader
        from .readers.txt import TxtReader
        from .readers.json import JSONReader
        from .readers.html import HTMLReader
        from .readers.excel import ExcelReader

        return {
            ".pdf": PyMuPDFReader(),
            ".csv": CSVReader(),
            ".docx": DocxReader(),
            ".txt": TxtReader(),
            ".json": JSONReader(),
            ".html": HTMLReader(),
            ".xlsx": ExcelReader(),
            ".xls": ExcelReader(),
        }

    def _get_all_file_paths(
        self,
        file_or_folder_paths: list[str | Path],
        should_splits: list[bool],
        extra_infos: list[dict],
    ) -> list[Path]:
        all_file_paths = []
        new_should_splits = []
        new_extra_infos = []

        for path, should_split, extra_info in zip(
            file_or_folder_paths, should_splits, extra_infos
        ):
            if isinstance(path, str):
                path = Path(path)
            if isinstance(path, Path):
                if path.is_dir():
                    files = [
                        file
                        for file in path.glob("**/*")
                        if file.is_file() and file.suffix in self.processors
                    ]
                    all_file_paths.extend(files)
                    new_should_splits.extend([should_split] * len(files))
                    new_extra_infos.extend([extra_info] * len(files))
                elif path.is_file() and path.suffix in self.processors:
                    all_file_paths.append(path)
                    new_should_splits.append(should_split)
                    new_extra_infos.append(extra_info)
                else:
                    warnings.warn(f"Invalid file_paths: {path}")
        return (
            [str(file) for file in all_file_paths],
            new_should_splits,
            new_extra_infos,
        )

    def load_data(
        self,
        file_or_folder_paths: list[str | Path],
        should_splits: list[bool] | None = None,
        extra_infos: list[dict] | None = None,
        num_workers: int = 1,
    ) -> list[Document]:
        logger.info(f"Processing {len(file_or_folder_paths)} paths ...")

        should_splits = should_splits or [True] * len(file_or_folder_paths)
        extra_infos = extra_infos or [None] * len(file_or_folder_paths)

        assert len(file_or_folder_paths) == len(should_splits), (
            "file_or_folder_paths and should_splits must have the same length."
        )

        assert len(file_or_folder_paths) == len(extra_infos), (
            "file_or_folder_paths and extra_infos must have the same length."
        )

        file_paths, should_splits, extra_infos = self._get_all_file_paths(
            file_or_folder_paths, should_splits, extra_infos
        )

        args_list = [
            (file_paths[i], should_splits[i], extra_infos[i], self.processors)
            for i in range(len(file_paths))
        ]

        documents = []

        if num_workers == 1:
            for args in args_list:
                documents.extend(process_file_worker(args))
        else:
            with Pool(num_workers) as pool:
                results = pool.map(process_file_worker, args_list)
                for result in results:
                    documents.extend(result)

        logger.info(f"Get {len(documents)} documents.")
        return documents

    async def aload_data(
        self,
        file_or_folder_paths: list[str | Path],
        should_splits: list[bool] | None = None,
        extra_infos: list[dict] | None = None,
        max_workers: int = 1,
    ) -> list[Document]:
        """
        Asynchronous version of load_data.
        """
        logger.info(f"Processing {len(file_or_folder_paths)} files asynchronously ...")

        should_splits = should_splits or [True] * len(file_or_folder_paths)
        extra_infos = extra_infos or [None] * len(file_or_folder_paths)

        assert len(file_or_folder_paths) == len(should_splits), (
            "file_or_folder_paths and should_splits must have the same length."
        )

        assert len(file_or_folder_paths) == len(extra_infos), (
            "file_or_folder_paths and extra_infos must have the same length."
        )

        file_paths, should_splits, extra_infos = self._get_all_file_paths(
            file_or_folder_paths, should_splits, extra_infos
        )

        args_list = [
            (file_paths[i], should_splits[i], extra_infos[i], self.processors)
            for i in range(len(file_paths))
        ]

        documents = []

        if max_workers == 1:
            for args in args_list:
                documents.extend(await process_file_worker_async(args))
        else:
            with Pool(max_workers) as pool:
                results = await pool.map(process_file_worker_async, args_list)
                for result in results:
                    documents.extend(result)

        logger.info(f"Get {len(documents)} documents.")
        return documents



import asyncio
from dataclasses import dataclass
from abc import ABC, abstractmethod
from rag_colls.types.retriever import (
    RetrieverResult,
    RetrieverQueryType,
    RetrieverIngestInput,
)


@dataclass
class BaseRetrieverProvider(ABC):
    @abstractmethod
    def _test_connection(self):
        """
        Test the connection to the retriever provider.
        This method should be overridden by subclasses to implement the actual connection test.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    @abstractmethod
    def _add_documents(
        self,
        documents: list[RetrieverIngestInput],
        **kwargs,
    ):
        """
        Ingest documents into the retriever database.

        Args:
            documents (list[RetrieverIngestInput]): List of documents to ingest.
        """
        raise NotImplementedError(
            "The _add_documents method must be implemented in the subclass."
        )

    @abstractmethod
    def _search(self, query: RetrieverQueryType, **kwargs) -> list[RetrieverResult]:
        """
        Search for documents in the retriever provider.
        """
        raise NotImplementedError(
            "The _search method must be implemented in the subclass."
        )

    async def _asearch(
        self,
        query: RetrieverQueryType,
        **kwargs,
    ) -> list[RetrieverResult]:
        """
        Asynchronous search for documents in the retriever provider.
        """
        return await asyncio.to_thread(self._search, query, **kwargs)

    def add_documents(
        self,
        documents: list[RetrieverIngestInput],
        **kwargs,
    ):
        """
        Ingest documents into the retriever database.

        Args:
            documents (list[RetrieverIngestInput]): List of documents to ingest.
        """
        return self._add_documents(documents, **kwargs)

    def search(
        self,
        query: RetrieverQueryType,
        **kwargs,
    ) -> list[RetrieverResult]:
        """
        Search for documents in the retriever provider.
        """
        return self._search(query=query, **kwargs)

    async def asearch(
        self,
        query: RetrieverQueryType,
        **kwargs,
    ) -> list[RetrieverResult]:
        """
        Asynchronous search for documents in the retriever provider.
        """
        return await self._asearch(query=query, **kwargs)



import asyncio
from pathlib import Path
from abc import ABC, abstractmethod

from rag_colls.types.core.document import Document


class BaseReader(ABC):
    @abstractmethod
    def _load_data(
        self,
        file_path: str | Path,
        should_split: bool = True,
        extra_info: dict | None = None,
    ) -> list[Document]:
        """
        Loads data from the specified file path and returns a list of Document objects.

        Args:
            file_path (str | Path): The path to the file to be loaded.
            should_split (bool): Whether to split the data into smaller chunks.
            extra_info (dict | None): Additional information to be passed to the loader.

        Returns:
            list[Document]: A list of Document objects.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")

    def __str__(self):
        return f"{self.__class__.__name__}"

    async def _aload_data(
        self,
        file_path: str | Path,
        should_split: bool = True,
        extra_info: dict | None = None,
    ) -> list[Document]:
        """
        Asynchronously loads data from the specified file path and returns a list of Document objects.

        Args:
            file_path (str | Path): The path to the file to be loaded.
            extra_info (dict | None): Additional information to be passed to the loader.

        Returns:
            list[Document]: A list of Document objects.
        """
        return await asyncio.to_thread(
            self._load_data, file_path, should_split, extra_info
        )

    def load_data(
        self,
        file_path: str | Path,
        should_split: bool = True,
        extra_info: dict | None = None,
    ) -> list[Document]:
        """
        Loads data from the specified file path and returns a list of Document objects.

        Args:
            file_path (str | Path): The path to the file to be loaded.
            should_split (bool): Whether to split the data into smaller chunks.
            extra_info (dict | None): Additional information to be passed to the loader.

        Returns:
            list[Document]: A list of Document objects.
        """
        documents = self._load_data(file_path, should_split, extra_info)

        for doc in documents:
            assert doc.metadata.get("should_split") is not None, (
                f"Document metadata should contain 'should_split' key. Got: {doc.metadata}"
            )

        return documents

    async def aload_data(
        self,
        file_path: str | Path,
        should_split: bool = True,
        extra_info: dict | None = None,
    ) -> list[Document]:
        """
        Asynchronously loads data from the specified file path and returns a list of Document objects.

        Args:
            file_path (str | Path): The path to the file to be loaded.
            should_split (bool): Whether to split the data into smaller chunks.
            extra_info (dict | None): Additional information to be passed to the loader.

        Returns:
            list[Document]: A list of Document objects.
        """
        documents = await self._aload_data(file_path, should_split, extra_info)

        for doc in documents:
            assert doc.metadata.get("should_split") is not None, (
                f"Document metadata should contain 'should_split' key. Got: {doc.metadata}"
            )

        return documents
